import logging

def log_if_logger(*args, **kwargs) -> None:

    # use supplied logger if supplied, or root if it has been setup, otherwise no logger, just exit
    if 'logger' in kwargs and type(kwargs['logger']) in [logging.RootLogger, logging.Logger]: 
        errlogger = kwargs['logger']
    else: 
        try:
            if logging.getLogger().hasHandlers():  
                errlogger = logging.getLogger()
            else: 
                return None
        except:
            return None 

    # use message if supplied in kwargs, or as pos0 in args
    if 'message' in kwargs: 
        msg = kwargs['message']  
    else: 
        msg = 'No message supplied' if len(args) == 0 else str(args[0])    

    # log message
    errlogger.error(msg)
    return None


class SxTBiscuitError(Exception):
    def __init__(self, *args: object, **kwargs) -> None:
        log_if_logger(*args, **kwargs)
        super().__init__(*args)

class SxTKeyEncodingError(Exception):
    def __init__(self, *args: object, **kwargs) -> None:
        log_if_logger(*args, **kwargs)
        super().__init__(*args)

class SxTArgumentError(Exception):
    def __init__(self, *args: object, **kwargs) -> None:
        log_if_logger(*args, **kwargs)
        super().__init__(*args)
    
class SxTFileContentError(Exception):
    def __init__(self, *args: object, **kwargs) -> None:
        log_if_logger(*args, **kwargs)
        super().__init__(*args)
    
class SxTQueryError(Exception):
    def __init__(self, *args: object, **kwargs) -> None:
        log_if_logger(*args, **kwargs)
        super().__init__(*args)

class SxTAuthenticationError(Exception):
    def __init__(self, *args: object, **kwargs) -> None:
        log_if_logger(*args, **kwargs)
        super().__init__(*args)

class SxTAPINotDefinedError(Exception):
    def __init__(self, *args: object, **kwargs) -> None:
        log_if_logger(*args, **kwargs)
        super().__init__(*args)


class SxTAPINotSuccessfulError(Exception):
    def __init__(self, *args: object, **kwargs) -> None:
        log_if_logger(*args, **kwargs)
        super().__init__(*args)


class SxTExceptions():
    SxTAuthenticationError = SxTAuthenticationError
    SxTQueryError = SxTQueryError
    SxTFileContentError = SxTFileContentError
    SxTArgumentError = SxTArgumentError
    SxTKeyEncodingError = SxTKeyEncodingError
    SxTBiscuitError = SxTBiscuitError
    SxTAPINotDefinedError = SxTAPINotDefinedError
    SxTAPINotSuccessfulError = SxTAPINotSuccessfulError


from enum import Enum

class SXTPermission(Enum):
    SELECT = 'dql_select'
    INSERT = 'dml_insert'
    UPDATE = 'dml_update'
    DELETE = 'dml_delete'
    MERGE  = 'dml_merge'
    CREATE = 'ddl_create'
    ALTER  = 'ddl_alter'
    DROP   = 'ddl_drop'
    ALL    = '*'
    def __str__(self) -> str:
        return super().__str__()
    

class SXTKeyEncodings(Enum):
    HEX = 'hex'
    BASE64 = 'base64'
    BYTES = 'bytes'
    def __str__(self) -> str:
        return super().__str__()
    
    
class SXTApiCallTypes(Enum):
    POST = 'post'
    GET = 'get'
    PUT = 'put'
    DELETE = 'delete'
    def __str__(self) -> str:
        return super().__str__()
    

class SXTSqlType(Enum):
    DDL = 'ddl'
    DML = 'dml'
    DQL = 'dql'
    def __str__(self) -> str:
        return super().__str__()
    

class SXTOutputFormat(Enum):
    JSON = 'json'
    CSV = 'csv'
    DATAFRAME = 'dataframe'
    PARQUET = 'parquet'
    def __str__(self) -> str:
        return super().__str__()
    
    
class SXTTableAccessType(Enum):
    PERMISSIONED = 'permissioned'
    PUBLIC_READ = 'public_read'
    PUBLIC_APPEND = 'public_append'
    PUBLIC_WRITE = 'public_write'
    def __str__(self) -> str:
        return super().__str__()
    

class SXTResourceType(Enum):
    UNDEFINED = 'undefined'
    TABLE = 'table_name'
    VIEW = 'view_name'
    MATERIALIZED_VIEW = 'matview_name'
    PARAMETERIZED_VIEW = 'parmview_name'
    KAFKA_STREAM = 'kafka_name'
    def __str__(self) -> str:
        return super().__str__()
    

class SXTDiscoveryScope(Enum):
    PRIVATE = 'private'
    SUBSCRIPTION = 'subscription'
    PUBLIC = 'public'
    ALL = 'all'
    def __str__(self) -> str:
        return super().__str__()
    



import logging, json
from pathlib import Path
from datetime import datetime
from biscuit_auth import KeyPair, PrivateKey, PublicKey, Authorizer, Biscuit, BiscuitBuilder, BlockBuilder, Rule, DataLogError
from .sxtexceptions import SxTArgumentError, SxTFileContentError, SxTBiscuitError, SxTKeyEncodingError
from .sxtenums import SXTPermission, SXTKeyEncodings
from .sxtkeymanager import SXTKeyManager



class SXTBiscuit():
    """Definition of a single biscuit."""

    logger: logging.Logger = None 
    domain: str = 'sxt'
    name: str = 'biscuit_name'
    key_manager: SXTKeyManager = None
    GRANT = SXTPermission
    ENCODINGS = SXTKeyEncodings
    __cap:dict = {'schema.resource':['permission1', 'permission2']}
    __bt: str = ''
    __lastresource: str = ''
    __manualtoken:bool = False
    __parentbiscuit__:bool = False
    __defaultresource__:str = ''
    
    def __init__(self, name:str = '', private_key: str = None, new_keypair: bool = False, 
                 from_file: Path = None, logger:logging.Logger = None, 
                 biscuit_token:str = None, default_resource:str = None) -> None:
        if logger: 
            self.logger = logger 
        else: 
            self.logger = logging.getLogger()
            self.logger.setLevel(logging.INFO)
            if len(self.logger.handlers) == 0: 
                self.logger.addHandler( logging.StreamHandler() )
        self.logger.info('-'*30 + '\nNew SXT Biscuit initiated')
        self.key_manager = SXTKeyManager(logger=self.logger, encoding=SXTKeyEncodings.BASE64)
        if new_keypair: self.key_manager.new_keypair()
        if private_key: self.private_key = private_key
        self.__cap = {}
        if name: self.name = name 
        if default_resource: self.__defaultresource__ = default_resource
        if from_file and Path(from_file).exists: self.load(from_file)
        if biscuit_token: 
            self.__manualtoken = True 
            self.__bt = biscuit_token
            self.logger.info('manual biscuit token accepted as-is, not calculated or verified.')
            

    def __str__(self):
        return '\n'.join([f"{str(n).rjust(25)}: {v}" for n,v in  dict(self.to_json(True, True)).items()])
    
    def __repr__(self):
        return '\n'.join([f"{str(n).rjust(25)}: {v}" for n,v in  dict(self.to_json(False, True)).items()])

    def __len__(self):
        return 1

    @property
    def biscuit_text(self):
        b = []
        for resource, permissions in self.__cap.items():
            for permission in sorted(permissions):
                b.append(f'{self.domain}:capability("{permission}", "{str(resource).lower()}");')
        return '\n'.join(b)

    @property
    def biscuit_token(self) ->str:
        if self.__manualtoken: return self.__bt
        if not self.private_key or not self.biscuit_text: 
            self.__bt = ''
            return ''
        if not self.__bt: self.__bt = self.regenerate_biscuit_token()
        return self.__bt
    
    @property
    def biscuit_json(self) -> dict:
        return self.__cap

    @property
    def private_key(self) ->str :
        return self.key_manager.private_key
    @private_key.setter
    def private_key(self, value):
        self.key_manager.private_key = value 
        self.__bt = ''

    @property
    def public_key(self) ->str :
        return self.key_manager.public_key
    @public_key.setter
    def public_key(self, value):
        self.key_manager.public_key = value 

    @property
    def encoding(self) ->str :
        return self.key_manager.encoding
    @encoding.setter
    def encoding(self, value):
        self.key_manager.encoding = value     

    @property
    def default_resource(self) ->str: 
        return str(self.__defaultresource__)
    @default_resource.setter
    def default_resource(self, value):
        if type(value)==str: 
            self.__defaultresource__ = value
        else:
            raise ValueError('default_resource must be set to a string.')


    def new_keypair(self) -> dict:
        return self.key_manager.new_keypair()


    def regenerate_biscuit_token(self) -> dict:
        """--------------------
        Regenerates the biscuit_token from class.biscuit_text and class.private_key. 

        For object consistency, this only leverages the class objects, so there are no arguments.  
        To build the biscuit_text, clear_capabilities() and then add_capability(), or if you want to import 
        an existing datalog file, you can load capabilities_from_text(). This function will error without 
        a valid private_key and biscuit_text.

        Args: 
            None

        Returns: 
            str: biscuit_token in base64 format. 
        """
        if not self.private_key:
            raise SxTArgumentError("Private Key is required to create a biscuit", logger=self.logger)
        
        biscuit_text = self.biscuit_text 
        if not biscuit_text:
            raise SxTArgumentError('Biscuit_Text is required to create a biscuit. Try to add_capability() and inspect biscuit_text to verify.', logger=self.logger)

        try:
            private_key_obj = PrivateKey.from_hex(self.key_manager.private_key_to(SXTKeyEncodings.HEX))
            biscuit = BiscuitBuilder(self.biscuit_text).build(private_key_obj)
            return biscuit.to_base64() 
        except DataLogError as ex:
            errmsg = ex
        raise SxTBiscuitError(errmsg, logger=self.logger)


    def validate_biscuit(self, biscuit_base64:str, public_key = None) -> str:
        if not public_key: public_key = self.public_key
        public_key = self.convert_key(public_key, self.get_encoding_type(public_key), SXTKeyEncodings.HEX )
        try:
            return Biscuit.from_base64( data=biscuit_base64, root=PublicKey.from_hex( public_key ))
        except Exception as ex:
            self.logger.error(ex)
        raise SxTBiscuitError('Biscuit not validated with Public Key', logger=self.logger)


    def add_capability(self, resource:str, *permissions):
        """--------------------
        Adds a capability to the existing biscuit structure. 

        Args:
            resource (str): Resource (Schema.Resource) to which permissions are applied.
            permission (*): Any number of SXTPermission enums to GRANT to the resource. 

        Returns:
            int: Number of items added (excluding duplicates)

        Examples:
            >>> bb = SXTBiscuitBuilder()
            >>> bb.add_capability("Schema.TableA", "SELECT")
            1
            >>> bb.add_capability("Schema.TableA", "INSERT")
            True
            >>> bb.add_capability("Schema.TableA", "INSERT")
            False
        """
        if not resource: resource = self.__defaultresource__
        if not resource: resource = self.__lastresource
        if not resource: raise KeyError('must define a resource (schema.object) on which to assign permissions')
        self.__lastresource = resource
        if resource not in self.__cap: self.__cap[resource] = []
        if 'ALL' in self.__cap[resource] or '*' in self.__cap[resource]:
            self.logger.warning('Cannot add other permissions to a biscuit containing ALL permissions.  Request disregarded.')
            return self.__isall__(resource)
        initial_count = len(self.__cap[resource])
        process_count = 0
        final_permissions = []
        for permission in permissions: 
            process_count += 1
            if type(permission) == list: 
                if 'ALL' in permission: return self.__isall__(resource)
                final_permissions += list(permission)
                process_count += len(list(permission))-1
            else:
                if permission.name == 'ALL': return self.__isall__(resource)
                final_permissions.append(permission)
        self.__cap[resource] += [p.value for p in final_permissions]
        self.__cap[resource] = list(set(self.__cap[resource]))
        self.__bt = ''
        added_count = len(self.__cap[resource]) - initial_count
        self.logger.info(f'Added {added_count} permissions, from total {process_count} submitted ({process_count - added_count} duplicates)')
        return added_count

    def __isall__(self, resource):
        total_count = len(self.__cap[resource])
        self.__cap[resource] = ['*']
        self.logger.info(f'Added ALL permissions, replacing a total of {total_count} other permissions.')
        return 1
        

    def capabilities_from_text(self, biscuit_text:str) -> None:
        """--------------------
        Loads text into biscuit capabilities, for example, loading a datalog file directly.

        Args: 
            biscuit_text (str): Text to compile into capabilities.

        Results:
            str: biscuit_text that has been digested and re-processed.

        """
        biscuit_lines = str(biscuit_text).strip().split('\n')
        caps = {}
        self.logger.debug(f'Translating supplied text into biscuit capabilities...')
        for line in biscuit_lines:
            if line.strip().startswith(f'{self.domain}:capability'):
                c = line.split('"')
                if len(c) <5: 
                    raise SxTArgumentError('biscuit_text capabilities must have format  domain:capability("PERMISSION", "RESOURCE");')
                p = c[1] # permission
                r = c[3] # resource
                if r not in caps: caps[r] = []
                caps[r].append(p)
        for r in list(caps.keys()):
            caps[r] = list(set(caps[r]))
        self.__cap = caps
        self.__bt = ''
        self.logger.debug(f'Successfully translated biscuit_text to biscuit capabilities objects.')
        return None


    def capabilities_from_token(self, biscuit_token:str) -> None:
        raise NotImplementedError('Not implemented yet. Please check back later.')


    def clear_capabilities(self):
        """Clears all existing capabilities from the biscuit structure"""
        self.__cap = {}
        self.__bt = ''
        self.logger.debug('Clearing all biscuit capabilities')
        return None


    def to_json(self, mask_private_key:bool = True, add_tabs_to_biscuit_text:bool = False):
        """Exports content of biscuit to a json format.  WARNING, this can include private key."""
        tab = '\t' if add_tabs_to_biscuit_text else ''
        rtn = { 'private_key': getattr(self, 'private_key')[:6]+'...' if mask_private_key else getattr(self, 'private_key')
               ,'public_key' : getattr(self, 'public_key' )
               ,'biscuit_capabilities' : self.__cap
               ,'biscuit_token': self.biscuit_token
               ,'biscuit_text': f'\n{tab}' + str(self.biscuit_text).replace('\n',f'\n{tab}')
               }
        self.logger.debug(f'Translating object data to json')
        return dict(rtn)


    def save(self, filepath: Path = 'biscuits/biscuit_{resource}_{date}_{time}.json', overwrite:bool = False, resource:str = None) -> Path:
        """--------------------
        Saves biscuit information to a json file.

        The filepath will accept three different placholder texts: {resource}, {date}, and {time}.  
        This allows caller to easily create dynamically named biscuit files, reducing the likelihood of 
        overwriting biscuit files and thus losing keys.   It is best practice to leave overwrite to False
        and use placeholders to save different files, removing older save files only after validation
        the keys are not needed anymore.  

        Args:
            filepath (Path): Full file path which to save, with placeholders allowed.
            overwrite (bool): If True, will overwrite file if exists
            resource (str): Optional resource name for placeholder in filepath.  Defaults to last resource set in add_capability().
        
        Results:
            Path: Same as filepath, if successful
        """
        # TODO: add string.replace for resource, date, time
        filepath = Path(filepath).resolve()
        # do placeholder replacements
        if not resource: resource = self.__lastresource
        date = datetime.now().strftime('%Y%m%d')
        time = datetime.now().strftime('%H%M%S')
        filepath = Path(str(filepath).replace('{resource}', resource).replace('{date}',date).replace('{time}',time))
        if filepath.exists() and not overwrite: 
            raise FileExistsError(f'{filepath} already exists.  Set overwrite = True to overwrite automatically.')

        filepath.parent.mkdir(parents=True, exist_ok=True)
        self.logger.debug(f'Opening file to write: {filepath}...')
        with open(filepath, 'w') as fh:
            fh.write( json.dumps(self.to_json(False)).replace('\t','') )
        self.logger.debug(f'Data written to file.')
        return filepath
    

    def load(self, filepath: Path, resource:str = None, date:str = None, time:str = None) -> dict:
        """--------------------
        Loads a biscuit from correctly formated JSON file.

        The filepath will accept three different placholder texts: {resource}, {date}, and {time}.  
        This allows caller to easily create (and load) dynamically named biscuit files, reducing the 
        likelihood of overwriting biscuit files and thus losing keys.  It is best practice to leave 
        overwrite to False and use placeholders to save different files, removing older save files 
        only after validation the keys are not needed anymore.

        Args:
            filepath (Path): Full file path which to load from
            resource (str): Optional resource name for placeholder in filepath.  Defaults to last resource set in add_capability().
            date (str): Optional integer-only date (yyyymmdd) for placeholder in file path. Defaults to current date.
            time (str): Optional integer-only time (hhmmss) for placeholder in file path. Defaults to current time.

        Results:
            Path: Same as filepath, if successful
        """
        self.logger.info('Attempting to load biscuit definition from file...')
        filepath = Path(filepath).resolve()
        # do placeholder replacements
        if not resource: resource = self.__lastresource
        if not date: date = datetime.now().strftime('%Y%m%d')
        if not time: time = datetime.now().strftime('%H%M%S')
        filepath = Path(str(filepath).replace('{resource}', resource).replace('{date}',date).replace('{time}',time))
        # TODO: allow option for 'Most Recent" for date and time, which can look thru
        # the parent directory and find the file with the most recent {date} / {time}.
        # This will hopefully promote behavior of keeping history of keys, in case 
        # they're needed 
        
        if not filepath.exists: 
            raise FileNotFoundError(f'{filepath} not found.')
        try:
            self.logger.debug(f'Opening file: {filepath}...')
            with open(filepath, 'r') as fh:
                content = json.loads(fh.read())
            if 'private_key' not in content or 'biscuit_text' not in content: 
                raise SxTFileContentError 
        except (SxTFileContentError, json.JSONDecodeError):
              self.logger.error(f'File not loaded due to missing, malformed content, or simply unable to load JSON: \n{filepath}')
              return None
        try:
            new_key_encoding = self.key_manager.get_encoding_type(content['private_key'])
            new_private_key =  self.key_manager.convert_key(content['private_key'], new_key_encoding, SXTKeyEncodings.BYTES)
        except SxTArgumentError: 
            return None
        
        # Assign last, after all validation.  public_key, biscuit_text, biscuit_token all recalculate automatically.
        self.logger.info(f'File opened and parsed, loading data into current object.')
        self.capabilities_from_text(content['biscuit_text'])
        self.private_key = new_private_key
        return content
            





if __name__ == '__main__':
    
    print('\n', '-=-='*10, '\n' )

    # BASIC USAGE
    bis = SXTBiscuit(name='my first biscuit', new_keypair=True)
    bis.add_capability('schema.SomeTable',    bis.GRANT.SELECT, bis.GRANT.INSERT, bis.GRANT.UPDATE)
    bis.add_capability('schema.AnotherTable', bis.GRANT.SELECT)
    print( bis.biscuit_token )
    # check out https://www.biscuitsec.org/ for external validation.

     
    # Permissions can be supplied as individual items, a list of items, or both
    bis.clear_capabilities()
    bis.add_capability('schema.SomeTable', bis.GRANT.SELECT, [bis.GRANT.UPDATE, bis.GRANT.INSERT, bis.GRANT.DELETE], bis.GRANT.MERGE) 
    
    # Permissions will also deduplicate themselves
    bis.add_capability('schema.SomeTable', bis.GRANT.SELECT, [bis.GRANT.CREATE, bis.GRANT.INSERT, bis.GRANT.DROP], bis.GRANT.ALTER) 
    bis.add_capability('schema.SomeTable', bis.GRANT.MERGE, bis.GRANT.DELETE, bis.GRANT.SELECT, bis.GRANT.CREATE) 
    print( bis )

    # Resources with no permissions never make it into the biscuit
    bis.clear_capabilities()
    bis.add_capability('schema.NoPermissions')
    print( f'-->{bis.biscuit_text}<--' )
    print( f'-->{bis.biscuit_token}<--' )

    # You can add ALL permissions at once
    bis.clear_capabilities()
    bis.add_capability('schema.SomeTable', bis.GRANT.ALL, bis.GRANT.SELECT )
    print( bis )

    # To build a "wildcard" biscuit (although not recommended beyond testing)
    bis.clear_capabilities()
    bis.add_capability('*', bis.GRANT.ALL)
    print( bis )

    # Note, assigning ALL to permissions will remove all other permissions
    bis.clear_capabilities()
    bis.add_capability('*', bis.GRANT.SELECT, bis.GRANT.INSERT, bis.GRANT.DELETE)
    bis.add_capability('*', bis.GRANT.ALL)
    print( bis )

    # ALL biscuits will also prevent assignment of other permissions - see WARNING
    bis.add_capability('*', bis.GRANT.SELECT)
    print( bis )


    # Printing biscuit object as string (__str__) obscures the private key.
    # to print the full private key, print the representation instead.
    print( repr( bis ) )


    # Save biscuit information to disk, so keys are not lost
    bis.clear_capabilities()
    bis.add_capability('schema.SomeTable', bis.GRANT.SELECT, [bis.GRANT.UPDATE, bis.GRANT.INSERT, bis.GRANT.DELETE], bis.GRANT.MERGE) 
    save_file = './biscuits/biscuit_{resource}_{date}.json'
    print( bis.save(save_file, overwrite = True) )

    # load new biscuit object from saved json 
    bis2 = SXTBiscuit(logger=bis.logger, from_file = './biscuits/biscuit_schema.SomeTable_{date}.json')
    print( bis2 )

    # or 
    bis3 = SXTBiscuit(logger=bis.logger)
    bis3.load( './biscuits/biscuit_{resource}_{date}.json', resource='schema.SomeTable' )
    print( bis3 )

    # Practically speaking, you'll want to guard against losing keys.
    # Using the default filename will help.
    bis3.save(resource='schema.SomeTable')


    # Return a json object with information
    print( bis.to_json() )
    pass



import os, sys, pytest, pandas, random
from pathlib import Path

# load local copy of libraries
sys.path.append(str( Path(Path(__file__).parents[1] / 'src').resolve() ))
from spaceandtime.spaceandtime import SpaceAndTime
from spaceandtime.spaceandtime import SXTUser
from spaceandtime.sxtkeymanager import SXTKeyManager
from spaceandtime.sxtbiscuits import SXTBiscuit
from spaceandtime.sxtexceptions import *  # only contains exceptions prefixed with "SXT"
API_URL = 'https://api.spaceandtime.dev'


def test_sxt_exceptions():
    # test common exceptions
    sxt = SpaceAndTime() 
    sxt.user.user_id = ''
    with pytest.raises(Exception) as e_info: sxt.authenticate()
    with pytest.raises(SxTArgumentError) as e_info: sxt.authenticate()

    sxt = SpaceAndTime() 
    sxt.user.private_key = ''
    with pytest.raises(Exception) as e_info: sxt.authenticate()
    with pytest.raises(SxTArgumentError) as e_info: sxt.authenticate()

    with pytest.raises(SxTAuthenticationError) as errinfo:   raise SxTAuthenticationError('test message: SxTAuthenticationError')
    with pytest.raises(SxTQueryError) as errinfo:            raise SxTQueryError('test message: SxTQueryError')
    with pytest.raises(SxTFileContentError) as errinfo:      raise SxTFileContentError('test message: SxTFileContentError')
    with pytest.raises(SxTArgumentError) as errinfo:         raise SxTArgumentError('test message: SxTArgumentError')
    with pytest.raises(SxTKeyEncodingError) as errinfo:      raise SxTKeyEncodingError('test message: SxTKeyEncodingError')
    with pytest.raises(SxTBiscuitError) as errinfo:          raise SxTBiscuitError('test message: SxTBiscuitError')
    with pytest.raises(SxTAPINotDefinedError) as errinfo:    raise SxTAPINotDefinedError('test message: SxTAPINotDefinedError')
    with pytest.raises(SxTAPINotSuccessfulError) as errinfo: raise SxTAPINotSuccessfulError('test message: SxTAPINotSuccessfulError')


def test_sxt_wrapper():
    # pick up default .env file, with USERID="pySDK_tester"
    # note, that specific user must exist in .env this test to succeed.
    sxt = SpaceAndTime() 
    assert sxt.user.user_id == 'pySDK_tester'
    assert sxt.user.public_key == "Lu8fefHsAYxKfj7oaCx+Rtz7eNiPln6xbOxJJo0aIZQ="
    assert sxt.user.private_key[:6] == 'MeaW6J'

    assert len(sxt.access_token) == 0
    sxt.authenticate()
    assert len(sxt.access_token) > 0
    assert sxt.access_token[:4] == 'eyJ0' 
    assert sxt.user.user_id == 'pySDK_tester'

    success, data = sxt.execute_query('Select * from SXTLabs.Singularity limit 1')
    assert success
    assert data[0]['NAME'] == 'Singularity'
    assert type(data) == list
    assert type(data[0]) == dict

    # pick up specific .env file, with USERID="sxtlabs.crm.etl"
    # note, that specific user must exist for this test to succeed.
    sxt = SpaceAndTime(envfile_filepath='./.env_alt') 
    assert sxt.user.user_id == 'pySDK_tester2'
    assert sxt.user.public_key == "Lu8fefHsAYxKfj7oaCx+Rtz7eNiPln6xbOxJJo0aIZQ="
    assert sxt.user.private_key[:6] == 'MeaW6J'

    assert len(sxt.access_token) == 0
    sxt.authenticate()
    assert len(sxt.access_token) > 0

    success, data = sxt.execute_query('Select * from SXTLabs.Singularity limit 1')
    assert success
    assert data[0]['NAME'] == 'Singularity'

    # define specific outside of .env, in memory only
    userid = 'testuser_X_' + f"{random.randint(0,999999999999):012}"
    keypair = SXTKeyManager(new_keypair=True, encoding=sxt.ENCODINGS.BASE64)

    sxt = SpaceAndTime(api_url=API_URL, user_id=userid, user_private_key=keypair.private_key)
    assert sxt.user.user_id == userid
    assert sxt.user.private_key == keypair.private_key
    sxt.authenticate()

    success, data = sxt.execute_query('Select * from SXTLabs.Singularity limit 1')
    assert success
    assert data[0]['NAME'] == 'Singularity'

    
def test_sxt_user():
    # pick up default .env file, with USERID="pySDK_tester"
    # note, that specific user must be used for this test to succeed.
    sxt = None

    # UserA -- load .env file
    userA = SXTUser(dotenv_file='./.env', api_url=API_URL)
    assert userA.user_id == 'pySDK_tester'
    assert userA.public_key == "Lu8fefHsAYxKfj7oaCx+Rtz7eNiPln6xbOxJJo0aIZQ="
    assert userA.private_key[:6] == 'MeaW6J'

    assert len(userA.access_token) == 0
    userA.authenticate()
    assert len(userA.access_token) > 0
    assert userA.access_token[:4] == 'eyJ0' 
    assert userA.user_id == 'pySDK_tester'

    success, data = userA.execute_query("Select Name, 'A' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    assert data[0]['NAME'] == 'Singularity'
    assert data[0]['USERLETTER'] == 'A'

    # UserB -- load specific info
    userid = 'testuser_X_' + f"{random.randint(0,999999999999):012}"
    keypair = SXTKeyManager(new_keypair=True)
    userB = SXTUser(user_id=userid, user_private_key=keypair.private_key, api_url=API_URL, authenticate=True)
    success, data = userB.execute_query("Select Name, 'B' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    assert data[0]['NAME'] == 'Singularity'
    assert data[0]['USERLETTER'] == 'B'

    # Alternate querying with different users
    success, data = userA.execute_query("Select Name, 'A' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    success, data = userB.execute_query("Select Name, 'B' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    success, data = userA.execute_query("Select Name, 'A' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    success, data = userB.execute_query("Select Name, 'B' as UserLetter from SXTLabs.Singularity limit 1")
    assert success

    assert userA.user_id != userB.user_id
    assert userA.access_token != userB.access_token

    # this is just for backwards compatibility... prefer not to use:
    success, data = userA.execute_sql("Select Name, 'A' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    assert data[0]['NAME'] == 'Singularity'
    assert data[0]['USERLETTER'] == 'A'

    # test secret testuser functionality
    userC = SXTUser(testuser='C', user_private_key=keypair.private_key, api_url=API_URL, authenticate=True)
    success, data = userC.execute_sql("Select Name, 'C' as UserLetter from SXTLabs.Singularity limit 1")
    assert success
    assert data[0]['NAME'] == 'Singularity'
    assert data[0]['USERLETTER'] == 'C'

 
def test_execute_query():
    sxt = SpaceAndTime()
    sxt.authenticate()
    success, data = sxt.execute_query('Select * from SXTLabs.Singularity limit 1')
    assert success
    assert data[0]['NAME'] == 'Singularity'
    assert type(data) == list
    assert type(data[0]) == dict

    success, data = sxt.execute_query('Select * from SXTLabs.Singularity limit 1', 
                                      sql_type=sxt.SQLTYPE.DQL, resources=['SXTLabs.Singularity'],
                                      output_format = sxt.OUTPUT_FORMAT.PARQUET )
    assert success
    assert type(data) == bytes
 
    success, data = sxt.execute_query('Select * from SXTLabs.Singularity limit 1', 
                                      sql_type=sxt.SQLTYPE.DQL, resources=['SXTLabs.Singularity'],
                                      output_format = sxt.OUTPUT_FORMAT.DATAFRAME )
    assert success
    assert type(data) == pandas.DataFrame

    success, data = sxt.execute_query('Select * from SXTLabs.Singularity limit 1', 
                                      sql_type=sxt.SQLTYPE.DQL, resources=['SXTLabs.Singularity'],
                                      output_format = sxt.OUTPUT_FORMAT.CSV )
    assert success
    assert type(data) == list
    assert type(data[0]) == str # header
    assert type(data[1]) == str # data
    assert len(data) == 2 # header + 1 data row
    assert data[1].count(',') > 3


def test_discovery():
    sxt = SpaceAndTime()
    sxt.authenticate()

    # Schemas
    success, schemas = sxt.discovery_get_schemas(return_as=list)
    assert success
    assert type(schemas) == list 
    assert 'ETHEREUM' in schemas
    assert 'SXTDEMO' in schemas
    assert 'SXTLABS' in schemas

    success, schemas = sxt.discovery_get_schemas(return_as=dict)
    assert success
    assert type(schemas) == list
    assert type(schemas[0]) == dict

    success, schemas = sxt.discovery_get_schemas(return_as=str)
    assert success
    assert type(schemas) == str
    assert 'POLYGON,' in schemas
    assert schemas.count(',') >= 10

    success, schemas = sxt.discovery_get_schemas(scope = sxt.DISCOVERY_SCOPE.PRIVATE)
    assert success
    assert schemas.count(',') == 0  # no such thing right now

    # Tables
    success, tables = sxt.discovery_get_tables('SXTLabs', scope = sxt.DISCOVERY_SCOPE.PRIVATE, return_as=list)
    assert success
    assert 'SXTLABS.CRM_ACCOUNTS' in tables
    assert len(tables) >=10

    success, tables = sxt.discovery_get_tables('SXTLabs', search_pattern='CRM_Cosell', scope = sxt.DISCOVERY_SCOPE.PRIVATE, return_as=list)
    assert success
    assert 'SXTLABS.CRM_COSELL_AGREEMENTS' in tables
    assert len(tables) <=10

    # Columns
    success, columns = sxt.discovery_get_table_columns('POLYGON', 'BLOCKS', return_as=list)
    assert success
    assert 'TIME_STAMP' in columns
    assert len(columns) > 5
    
    success, columns = sxt.discovery_get_table_columns('POLYGON', 'BLOCKS', search_pattern='BLOCK', return_as=dict)
    assert success
    assert 'BLOCK_NUMBER' in columns.keys()
    assert len(columns) < 5
    
    success, columns = sxt.discovery_get_table_columns('SXTLABS', 'CRM_ACCOUNTS') # defaults to json, aka list-of-dicts
    assert success
    assert 'CREATED_TIME' in [c['column'] for c in columns]
    assert 'ACCOUNT_NAME' in [c['column'] for c in columns]
    assert len(columns) > 15


if __name__ == '__main__':
    test_sxt_exceptions()
    test_sxt_wrapper()
    test_sxt_user()
    test_execute_query()
    test_discovery()
    pass 


import requests, logging, json
from pathlib import Path
from .sxtenums import SXTApiCallTypes
from .sxtexceptions import SxTArgumentError, SxTAPINotDefinedError
from .sxtbiscuits import SXTBiscuit


class SXTBaseAPI():
    __au__:str = None 
    access_token = ''
    logger: logging.Logger
    network_calls_enabled:bool = True
    standard_headers = {
                    "accept": "application/json",
                    "content-type": "application/json"
                    }
    versions = {}
    APICALLTYPE = SXTApiCallTypes


    def __init__(self, access_token:str = '', logger:logging.Logger = None) -> None:
        if logger: 
            self.logger = logger 
        else: 
            self.logger = logging.getLogger()
            self.logger.setLevel(logging.INFO)
            if len(self.logger.handlers) == 0: 
                self.logger.addHandler( logging.StreamHandler() )

        apiversionfile = Path(Path(__file__).resolve().parent / 'apiversions.json')
        self.access_token = access_token
        with open(apiversionfile,'r') as fh:
            content = fh.read()
        self.versions = json.loads(content)
        


    @property
    def api_url(self):
        return self.__au__ if self.__au__ else 'https://api.spaceandtime.dev' # default 
    @api_url.setter
    def api_url(self, value):
        self.__au__ = value
        
    def prep_biscuits(self, biscuits=[]) -> list:
        """--------------------
        Accepts biscuits in various data types, and returns a list of biscuit_tokens as strings (list of str).  
        Primary use-case is class-internal.

        Args: 
            biscuits (list | str | SXTBiscuit): biscuit_tokens as a list, str, or SXTBiscuit type. 

        Returns: 
            list: biscuit_tokens as a list.

        Examples:
            >>> sxt = SpaceAndTime()
            >>> biscuits = sxt.user.base_api.prep_biscuits(['a',['b','c'], 'd'])
            >>> biscuits == ['a', 'b', 'c', 'd']
            True
        """
        if   biscuits == None or len(biscuits) == 0:
            return [] 
        elif type(biscuits) == str:
            return [biscuits]
        elif type(biscuits) == SXTBiscuit:  
            return [biscuits.biscuit_token]
        elif type(biscuits) == list:
            rtn=[]
            for biscuit in biscuits:
                rtn = rtn + self.prep_biscuits(biscuit)
            return rtn 
        else:
            self.logger.warning(f"""Biscuit provided was an unexpected type: {type(biscuits)}
                                Type must be one of [ str | list | SXTBiscuit object | None ]
                                Ingnoring this biscuit entry. Biscuit value provided:
                                {biscuits}""")
            return []


    def prep_sql(self, sql_text:str) -> str:
        """-------------------
        Cleans and prepares sql_text for transmission and execution on-network.

        Args: 
            sql_text (str): SQL text to prepare.

        Returns:
            sql: slightly modified / cleansed SQL text

        Examples:
            >>> api = SXTBaseAPI()
            >>> sql = "Select 'complex \nstring   ' as A \n   \t from \n\t TableName  \n Where    A=1;"
            >>> newsql = api.prep_sql(sql)
            >>> newsql == "Select 'complex \nstring   ' as A from TableName Where A=1"
            True
        """
        if sql_text == None or len(sql_text.strip()) == 0: return ''
        insinglequote = False
        indoublequote = False 
        rtn = []
        char = prevchar = ''
        for char in list(sql_text.strip()):

            # escape anything in quotes
            if   char == "'": insinglequote = not insinglequote
            elif char == '"': indoublequote = not indoublequote
            if insinglequote or indoublequote:
                rtn.append(char)
                prevchar = ''
                continue 

            # replace newlines and tabs with spaces
            if char in ['\n', '\t']: char = ' '
            
            # remove double-spaces
            if char == ' ' and prevchar == ' ': continue      

            rtn.append(char)
            prevchar = char

        # remove ; if last character
        if char == ';': rtn = rtn[:-1]
        return str(''.join(rtn)).strip()
            
    
    def call_api(self, endpoint: str, 
                 auth_header:bool = True, 
                 request_type:str = SXTApiCallTypes.POST, 
                 header_parms: dict = {}, 
                 data_parms: dict = {}, 
                 query_parms: dict = {}, 
                 path_parms: dict = {},
                 endpoint_full_override_flag: bool = False ) -> tuple[bool, object]:
        """--------------------
        Generic function to call and return SxT API. 

        This is the base api execution function.  It can, but is not intended, to be used directly.
        Rather, it is wrapped by other api-specific functions, to isolate api call differences
        from the actual api execution, which can all be the same. 

        Args:
            endpoint (str): URL endpoint, after the version. Final structure is: [api_url/version/endpoint] 
            request_type (SXTApiCallTypes): Type of request. [POST, GET, PUT, DELETE]
            auth_header (bool): flag indicator whether to append the Bearer token to the header. 
            header_parms: (dict): Name/Value pair to add to request header, except for bearer token. {Name: Value}
            query_parms: (dict): Name/value pairs to be added to the query string. {Name: Value}
            data_parms (dict): Dictionary to be used holistically for --data json object.
            path_parms (dict): Pattern to replace placeholders in URL. {Placeholder_in_URL: Replace_Value}
            endpoint_full_override_flag (str): If True, endpoint is used verbatium, rather than constructing version + endpoint. Will negate any querystring or path parms.

        Results:
            bool: Indicating request success
            json: Result of the API, expressed as a JSON object 
        """
        # Set these early, in case of timeout and they're not set by callfunc 
        txt = 'response.text not available - are you sure you have the correct API Endpoint?' 
        statuscode = 555
        response = {}

        # if network calls turned off, return fake data
        if not self.network_calls_enabled: return True, self.__fakedata__(endpoint)

        # internal function to simplify and unify error handling
        def __handle_errors__(txt, ex, statuscode, responseobject, loggerobject):
            loggerobject.error(txt)
            rtn = {'text':txt}
            rtn['error'] = str(ex)
            rtn['status_code'] = statuscode 
            rtn['response_object'] = responseobject
            return False, rtn

        # otherwise, go get real data
        try:
            # Header parms
            headers = {k:v for k,v in self.standard_headers.items()} # get new object
            if auth_header: headers['authorization'] = f'Bearer {self.access_token}'
            headers.update(header_parms)


            if endpoint_full_override_flag:
                url = endpoint
                self.logger.debug(f'API Call started for (custom) endpoint: {endpoint}')

            else:
                if endpoint not in self.versions.keys() and not endpoint_full_override_flag: 
                    raise SxTAPINotDefinedError("Endpoint not defined in API Lookup (apiversions.json). Please reach out to Space and Time for assistance. \nAs a work-around, you can try manually adding the endpoint to the SXTBaseAPI.versions dictionary.")
                version = self.versions[endpoint]
                self.logger.debug(f'API Call started for endpoint: {version}/{endpoint}')
            
                # Path parms
                for name, value in path_parms.items():
                    endpoint = endpoint.replace(f'{{{name}}}', value)
                
                # Query parms
                if query_parms !={}: 
                    endpoint = f'{endpoint}?' + '&'.join([f'{n}={v}' for n,v in query_parms.items()])

                # final URL
                url = f'{self.api_url}/{version}/{endpoint}'

            if request_type not in SXTApiCallTypes: 
                msg = f'request_type must be of type SXTApiCallTypes, not { type(request_type) }'
                raise SxTArgumentError(msg, logger=self.logger)
            
            match request_type:
                case SXTApiCallTypes.POST   : callfunc = requests.post
                case SXTApiCallTypes.GET    : callfunc = requests.get
                case SXTApiCallTypes.PUT    : callfunc = requests.put
                case SXTApiCallTypes.DELETE : callfunc = requests.delete
                case _: raise SxTArgumentError('Call type must be SXTApiCallTypes enum.', logger=self.logger)

            # Call API function as defined above
            response = callfunc(url=url, data=json.dumps(data_parms), headers=headers)
            txt = response.text
            statuscode = response.status_code
            response.raise_for_status()

            try:
                self.logger.debug('API return content type: ' + response.headers.get('content-type','') )
                rtn = response.json()
            except json.decoder.JSONDecodeError as ex:
                rtn = {'text':txt, 'status_code':statuscode}

            self.logger.debug(f'API call completed for endpoint: "{endpoint}" with result: {txt}')
            return True, rtn

        except requests.exceptions.RequestException as ex:
            return __handle_errors__(txt, ex, statuscode, response, self.logger)
        except SxTAPINotDefinedError as ex:
            return __handle_errors__(txt, ex, statuscode, response, self.logger)
        except Exception as ex:
            return __handle_errors__(txt, ex, statuscode, response, self.logger)        
        

    def __fakedata__(self, endpoint:str):
        if endpoint in ['sql','sql/dql']:
            rtn = [{'id':'1', 'str':'a','this_record':'is a test'}]
            rtn.append( {'id':'2', 'str':'b','this_record':'is a test'} )
            rtn.append( {'id':'3', 'str':'c','this_record':'is a test'} )
            return rtn
        else:
            return {'authCode':'469867d9660b67f8aa12b2'
                        ,'accessToken':'eyJ0eXBlIjoiYWNjZXNzIiwia2lkIjUxNDVkYmQtZGNmYi00ZjI4LTg3NzItZjVmNjNlMzcwM2JlIiwiYWxnIjoiRVMyNTYifQ.eyJpYXQiOjE2OTczOTM1MDIsIm5iZiI6MTY5NzM5MzUwMiwiZXhwIjoxNjk3Mzk1MDAyLCJ0eXBlIjoiYWNjZXNzIiwidXNlciI6InN0ZXBoZW4iLCJzdWJzY3JpcHRpb24iOiIzMWNiMGI0Yi0xMjZlLTRlM2MtYTdhMS1lNWRmNDc4YTBjMDUiLCJzZXNzaW9uIjoiMzNiNGRhMzYxZjZiNTM3MjZlYmYyNzU4Iiwic3NuX2V4cCI6MTY5NzQ3OTkwMjMxNSwiaXRlcmF0aW9uIjoiNDEwY2YyZTgyYWZlODdmNDRiMzE4NDFiIn0.kpvrG-ro13P1YeMF6sjLh8wn1rO3jpCVeTrzhDe16ZmJu4ik1amcYz9uQff_XQcwBDrpnCeD5ZZ9mHqb_basew'
                        ,'refreshToken':'eyJ0eXBlIjoicmVmcmVzaCIsImtpZCITQ1ZGJkLWRjZmItNGYyOC04NzcyLWY1ZjYzZTM3MDNiZSIsImFsZyI6IkVTMjU2In0.eyJpYXQiOjE2OTczOTM1MDIsIm5iZiI6MTY5NzM5MzUwMiwiZXhwIjoxNjk3Mzk1MzAyLCJ0eXBlIjoicmVmcmVzaCIsInVzZXIiOiJzdGVwaGVuIiwic3Vic2NyaXB0aW9uIjoiMzFjYjBiNGItMTI2ZS00ZTNjLWE3YTEtZTVkZjQ3OGEwYzA1Iiwic2Vzc2lvbiI6IjMzYjRkYTM2MWY2YjUzNzI2ZWJmMjc1OCIsInNzbl9leHAiOjE2OTc0Nzk5MDIzMTUsIml0ZXJhdGlvbiI6IjQxMGNmMmU4MmFmZTg3ZjQ0YjMxODQxYiJ9.3vVYpTGBjXIejlaacaZOh_59O9ETfbvTCWvldoi0ojyXTRkTmENVpQRbw7av7yMM2jA7SRdEPQGGjYmThCfk9w'
                        ,'accessTokenExpires':1973950023160
                        ,'refreshTokenExpires':1973953023160
                        }


    def get_auth_challenge_token(self, user_id:str, prefix:str = None, joincode:str = None):
        """--------------------
        (alias) Calls and returns data from API: auth/code, which issues a random challenge token to be signed as part of the authentication workflow.
        
        Args: 
            user_id (str): UserID to be authenticated
            prefix (str): (optional) The message prefix for signature verification (used for improved front-end UX).
            joincode (str): (optional) Joincode if creating a new user within an existing subscription. 

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        return self.auth_code(user_id, prefix, joincode)
    

    def auth_code(self, user_id:str, prefix:str = None) -> tuple[bool, object]:
        """--------------------
        Calls and returns data from API: auth/code, which issues a random challenge token to be signed as part of the authentication workflow.
        
        Args: 
            user_id (str): UserID to be authenticated
            prefix (str): (optional) The message prefix for signature verification (used for improved front-end UX).
            joincode (str): (optional) Joincode if creating a new user within an existing subscription. 

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        dataparms = {"userId": user_id}
        if prefix: dataparms["prefix"] = prefix
        success, rtn = self.call_api(endpoint = 'auth/code', auth_header = False, data_parms = dataparms)
        return success, rtn if success else [rtn]


    def gateway_proxy_add_existing_user(self, access_token:str) -> tuple[bool, object]:
        """-------------------- 
        Adds an authenticated user to the gateway proxy.  Fails if the user is not authenticated.

        Args: 
            user_id (str): UserID to be authenticated

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            dict: New Studio Password, or error message in a dict(json).
        """        
        endpoint = f'https://proxy.api.spaceandtime.dev/auth/add-existing?accessToken={access_token}'
        success, response = self.call_api(endpoint = endpoint, 
                                          auth_header = False, 
                                          endpoint_full_override_flag=True)
        self.logger.warning(f'add_user_to_gateway_proxy: {response}')
        if success and 'tempPassword' in response: response = response['tempPassword']
        return success, response



    def gateway_proxy_login (self, user_id:str, password:str) -> tuple[bool, object]:
        """-------------------- 
        Login to the gateway proxy, and return the session id.

        Args: 
            user_id (str): UserID to be authenticated
            password (str): Current, working password

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Gateway Proxy, including the session id, access token, etc.
        """        
        endpoint = 'https://proxy.api.spaceandtime.dev/auth/login'
        success, response = self.call_api(endpoint = endpoint, 
                                          auth_header = False, 
                                          data_parms = {"userId": user_id, "password": password},
                                          endpoint_full_override_flag=True)
        return success, response



    def gateway_proxy_change_password(self, user_id:str, old_password:str, new_password:str, session_id:str = None) -> tuple[bool, object]:
        """-------------------- 
        Logs into the gateway proxy and changes the user's password. Assuming the old_password still works, does not require network authentication.

        Args: 
            user_id (str): UserID to be authenticated
            old_password (str): Current, working password
            new_password (str): New password
            session_id (str): (optional) Session ID if already authenticated, otherwise this function will login and return authentication information as well.

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Gateway Proxy, as list or dict(json). 
        """        
        endpoint = 'https://proxy.api.spaceandtime.dev/auth/reset'
        if not session_id:
            success, login_response = self.gateway_proxy_login(user_id, old_password)
            if not success: 
                raise SxTArgumentError(f'Failed to log into gateway proxy: {login_response}')
            session_id = login_response['sessionId']
 
        success, response = self.call_api(endpoint = endpoint, 
                                          auth_header = False, 
                                          header_parms={"sid": session_id},
                                          data_parms = {"userId": user_id, "sid": session_id, 
                                                        "tempPassword": old_password, "newPassword": new_password},
                                          endpoint_full_override_flag=True)
        self.logger.warning(f'changed gateway proxy password: {response}')
        if login_response: response.update(login_response)
        response['password'] = new_password
        return success, response




    def auth_code_register(self, user_id:str, email:str, joincode:str = None, prefix:str = None) -> tuple[bool, object]:
        """--------------------
        Calls and returns data from API: auth/code-register, which issues a random challenge token to be signed as part of the authentication workflow, but also requires additional information with which to register a new user on the network.
        
        Args: 
            user_id (str): UserID to be authenticated
            email (str): Email address to validate new user
            joincode (str): (optional) Joincode if creating a new user within an existing subscription. 
            prefix (str): (optional) The message prefix for signature verification (used for improved front-end UX).

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        dataparms = {"userId": user_id, "email": email}
        if prefix: dataparms["joincode"] = joincode
        if prefix: dataparms["prefix"] = prefix

        success, rtn = self.call_api(endpoint = 'auth/code-register', auth_header = False, data_parms = dataparms)
        return success, rtn if success else [rtn]


    def get_access_token(self, user_id:str, challange_token:str, signed_challange_token:str='', public_key:str=None, keymanager:object=None, scheme:str = "ed25519"):
        """--------------------
        (alias) Calls and returns data from API: auth/token, which validates signed challenge token and provides new Access_Token and Refresh_Token. 
        Can optionally supply a keymanager object, instead of the public_key and signed_challenge_token.        
        
        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        return self.auth_token(user_id, challange_token, signed_challange_token, public_key, keymanager, scheme)


    def auth_token(self, user_id:str, challange_token:str, signed_challange_token:str='', public_key:str=None, keymanager:object=None, scheme:str = "ed25519"):
        """--------------------
        Calls and returns data from API: auth/token, which validates signed challenge token and provides new Access_Token and Refresh_Token. 
        Can optionally supply a keymanager object, instead of the public_key and signed_challenge_token. 
        
        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        if keymanager: 
            try:
                public_key = keymanager.public_key_to(keymanager.ENCODINGS.BASE64)
                signed_challange_token = keymanager.sign_message(challange_token)
            except Exception as ex:
                return False, {'error':'keymanager object must be of type SXTKeyManager, if supplied.'}

        dataparms = { "userId": user_id
                     ,"signature": signed_challange_token
                     ,"authCode": challange_token
                     ,"key": public_key 
                     ,"scheme": scheme}
        success, rtn = self.call_api(endpoint='auth/token', auth_header=False, data_parms=dataparms)
        return success, rtn if success else [rtn]


    def token_refresh(self, refresh_token:str):
        """--------------------
        Calls and returns data from API: auth/refresh, which accepts a Refresh_Token and provides a new Access_Token and Refresh_Token.        
        
        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        headers = { 'authorization': f'Bearer {refresh_token}' }
        success, rtn = self.call_api('auth/refresh', False, header_parms=headers)
        return success, rtn if success else [rtn]


    def auth_logout(self):
        """--------------------
        Calls and returns data from API: auth/logout, which invalidates Access_Token and Refresh_Token.        
        
        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        success, rtn = self.call_api('auth/logout', True)
        return success, rtn if success else [rtn]


    def auth_validtoken(self):
        """--------------------
        Calls and returns data from API: auth/validtoken, which returns information on a valid token.        
        
        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        success, rtn = self.call_api('auth/validtoken', True, SXTApiCallTypes.GET)
        return success, rtn if success else [rtn]
    

    def auth_idexists(self, user_id:str ):
        """--------------------
        Calls and returns data from API: auth/idexists, which returns True if the User_ID supplied exists, False if not.        
        
        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        success, rtn = self.call_api('auth/idexists/{id}', False, SXTApiCallTypes.GET, path_parms={'id':user_id})
        return success, rtn if success else [rtn]
    
    
    def auth_keys(self):
        """--------------------
        Calls and returns data from API: auth/keys (get), which returns all keys for a valid token.        
        
        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        success, rtn = self.call_api('auth/keys', True, SXTApiCallTypes.GET)
        return success, rtn if success else [rtn]


    def auth_addkey(self, user_id:str, public_key:str, challange_token:str, signed_challange_token:str, scheme:str = "ed25519"):
        """--------------------
        Calls and returns data from API: auth/keys (post), which adds a new key to the valid token. Requires similar challenge/sign/return as authentication.        
        
        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        dataparms = { "authCode": challange_token
                    ,"signature": signed_challange_token
                    ,"key": public_key
                    ,"scheme": scheme }
        success, rtn = self.call_api('auth/keys', True, SXTApiCallTypes.POST, data_parms=dataparms)
        return success, rtn if success else [rtn]


    def auth_addkey_challenge(self):
        """--------------------
        Request a challenge token from the Space and Time network, for authentication.

        (alias) Calls and returns data from API: auth/keys (get), which returns all keys for a valid token.        
        
        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        return self.auth_keys_code()
    

    def auth_keys_code(self):
        """--------------------
        Calls and returns data from API: auth/keys (get), which returns all keys for a valid token.        
        
        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        success, rtn = self.call_api('auth/keys/code', True)
        return success, rtn if success else [rtn]
    

    def sql_exec(self, sql_text:str, biscuits:list = None, app_name:str = None, validate:bool = False):
        """--------------------
        Executes a database statement/query of arbitrary type (DML, DDL, DQL), and returns a status or data.

        Calls and returns data from API: sql, which runs arbitrary SQL and returns records (if any).
        This api call undergoes one additional SQL parse step to interrogate the type and 
        affected tables / views, so is slightly less performant (by 50-100ms) than the type-specific 
        api calls, sql_ddl, sql_dml, sql_dql.  Normal human interaction will not be noticed, but
        if tuning for high-performance applications, consider using the correct typed call.

        Args:        
            sql_text (str): SQL query text to execute. Note, there is NO placeholder replacement.
            biscuits (list): (optional) List of biscuit tokens for permissioned tables. If only querying public tables, this is not needed.
            app_name (str): (optional) Name that will appear in querylog, used for bucketing workload.
            validate (bool): (optional) Perform an additional SQL validation in-parser, before database submission.

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        headers = { 'originApp': app_name } if app_name else {}
        sql_text = self.prep_sql(sql_text=sql_text)
        biscuit_tokens = self.prep_biscuits(biscuits)
        if type(biscuit_tokens) != list:  raise SxTArgumentError("sql_all requires parameter 'biscuits' to be a list of biscuit_tokens or SXTBiscuit objects.",  logger = self.logger)
        dataparms = {"sqlText": sql_text
                    ,"biscuits": biscuit_tokens
                    ,"validate": str(validate).lower() }
        success, rtn = self.call_api('sql', True, header_parms=headers, data_parms=dataparms)
        return success, rtn if success else [rtn]


    def sql_ddl(self, sql_text:str, biscuits:list = None, app_name:str = None):
        """--------------------
        Executes a database DDL statement, and returns status.

        Calls and returns data from API: sql/ddl, which runs arbitrary DDL for creating resources.        
        This will be slightly more performant than the generic sql_exec function, but requires a resource name.
        Biscuits are always required for DDL.

        Args: 
            sql_text (str): SQL query text to execute. Note, there is NO placeholder replacement.
            biscuits (list): (optional) List of biscuit tokens for permissioned tables. If only querying public tables, this is not needed.
            app_name (str): (optional) Name that will appear in querylog, used for bucketing workload.

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        headers = { 'originApp': app_name } if app_name else {}
        sql_text = self.prep_sql(sql_text=sql_text)
        biscuit_tokens = self.prep_biscuits(biscuits)
        if biscuit_tokens==[]:  raise SxTArgumentError("sql_ddl requires 'biscuits', none were provided.", logger = self.logger)
        dataparms = {"sqlText": sql_text
                    ,"biscuits": biscuit_tokens }
                    # ,"resources": [r for r in resources] }
        success, rtn = self.call_api('sql/ddl', True, header_parms=headers, data_parms=dataparms)
        return success, rtn if success else [rtn]


    def sql_dml(self, sql_text:str, resources:list, biscuits:list = None, app_name:str = None):
        """--------------------
        Executes a database DML statement, and returns status.

        Calls and returns data from API: sql/dml, which runs arbitrary DML for manipulating data. 
        This will be slightly more performant than the generic sql_exec function, but requires a resource name.
        Biscuits are required for any non-public-write tables.

        Args: 
            sql_text (str): SQL query text to execute. Note, there is NO placeholder replacement.
            resources (list): List of Resources ("schema.table_name") in the sql_text. 
            biscuits (list): (optional) List of biscuit tokens for permissioned tables. If only querying public tables, this is not needed.
            app_name (str): (optional) Name that will appear in querylog, used for bucketing workload.
        
        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        if type(resources) != list: resources = [resources]
        headers = { 'originApp': app_name } if app_name else {}
        sql_text = self.prep_sql(sql_text=sql_text)
        biscuit_tokens = self.prep_biscuits(biscuits)
        if type(biscuit_tokens) != list:  raise SxTArgumentError("sql_all requires parameter 'biscuits' to be a list of biscuit_tokens or SXTBiscuit objects.",  logger = self.logger)
        headers = { 'originApp': app_name } if app_name else {}
        dataparms = {"sqlText": sql_text
                    ,"biscuits": biscuit_tokens
                    ,"resources": [r for r in resources] }
        success, rtn = self.call_api('sql/dml', True, header_parms=headers, data_parms=dataparms)
        return success, rtn if success else [rtn]


    def sql_dql(self, sql_text:str, resources:list, biscuits:list = None, app_name:str = None):
        """--------------------
        Executes a database DQL / SQL query, and returns a dataset as a list of dictionaries.

        Calls and returns data from API: sql/dql, which runs arbitrary SELECT statements that return data.        
        This will be slightly more performant than the generic sql_exec function, but requires a resource name.
        Biscuits are required for any non-public tables.

        Args: 
            sql_text (str): SQL query text to execute. Note, there is NO placeholder replacement.
            resources (list): List of Resources ("schema.table_name") in the sql_text. 
            biscuits (list): (optional) List of biscuit tokens for permissioned tables. If only querying public tables, this is not needed.
            app_name (str): (optional) Name that will appear in querylog, used for bucketing workload.

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        if type(resources) != list: resources = [resources]
        headers = { 'originApp': app_name } if app_name else {}
        sql_text = self.prep_sql(sql_text=sql_text)
        biscuit_tokens = self.prep_biscuits(biscuits)
        if type(biscuit_tokens) != list:  raise SxTArgumentError("sql_all requires parameter 'biscuits' to be a list of biscuit_tokens or SXTBiscuit objects.",  logger = self.logger)
        dataparms = {"sqlText": sql_text
                    ,"biscuits": biscuit_tokens
                    ,"resources": [r for r in resources] }
        success, rtn = self.call_api('sql/dql', True, header_parms=headers, data_parms=dataparms)
        return success, rtn if success else [rtn]

    def sql_exec_tamperproof(self, sql_text:str, biscuits:list = None):
        """--------------------
        Executes a ZK tamperproof database statement/query, and returns a status or data plus a ZK verification code.

        Args:        
            sql_text (str): SQL query text to execute. Note, there is NO placeholder replacement.
            biscuits (list): (optional) List of biscuit tokens for permissioned tables. If only querying public tables, this is not needed.

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        sql_text = self.prep_sql(sql_text=sql_text)
        biscuit_tokens = self.prep_biscuits(biscuits)
        if type(biscuit_tokens) != list:  raise SxTArgumentError("sql_all requires parameter 'biscuits' to be a list of biscuit_tokens or SXTBiscuit objects.",  logger = self.logger)
        dataparms = {"sqlText": sql_text
                    ,"biscuits": biscuit_tokens }
        success, rtn = self.call_api('sql/tamperproof-query', True, data_parms=dataparms)
        return success, rtn if success else [rtn]


    def discovery_get_schemas(self, scope:str = 'ALL'):
        """--------------------
        Connects to the Space and Time network and returns all available schemas.
        
        Calls and returns data from API: discover/schema 

        Args:
            scope (SXTDiscoveryScope): (optional) Scope of objects to return: All, Public, Subscription, or Private. Defaults to SXTDiscoveryScope.ALL.

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list of dict. 
        """
        success, rtn = self.call_api('discover/schema',True, SXTApiCallTypes.GET, query_parms={'scope':scope})
        return success, (rtn if success else [rtn]) 
        

    def discovery_get_tables(self, schema:str = 'ETHEREUM', scope:str = 'ALL', search_pattern:str = None):
        """--------------------
        Connects to the Space and Time network and returns all available tables within a schema.

        Calls and returns data from API: discover/table         
        
        Args:
            schema (str): Schema name to search for tables.
            scope (SXTDiscoveryScope): (optional) Scope of objects to return: All, Public, Subscription, or Private. Defaults to SXTDiscoveryScope.ALL.
            search_pattern (str): (optional) Tablename pattern to match for inclusion into result set. Defaults to None / all tables.

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list of dict. 
        """
        version = 'v2' if 'discover/table' not in list(self.versions.keys()) else self.versions['discover/table'] 
        schema_or_namespace = 'namespace' if version=='v1' else 'schema'
        query_parms = {'scope':scope.upper(), schema_or_namespace:schema.upper()}
        if version != 'v1' and search_pattern: query_parms['searchPattern'] = search_pattern
        success, rtn = self.call_api('discover/table',True,  SXTApiCallTypes.GET, query_parms=query_parms)
        return success, (rtn if success else [rtn]) 


    def discovery_get_views(self, schema:str = 'ETHEREUM', scope:str = 'ALL', search_pattern:str = None):
        """--------------------
        Connects to the Space and Time network and returns all available tables within a schema.

        Calls and returns data from API: discover/table         
        
        Args:
            schema (str): Schema name to search for tables.
            scope (SXTDiscoveryScope): (optional) Scope of objects to return: All, Public, Subscription, or Private. Defaults to SXTDiscoveryScope.ALL.
            search_pattern (str): (optional) Tablename pattern to match for inclusion into result set. Defaults to None / all tables.

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list of dict. 
        """
        version = 'v2' if 'discover/view' not in list(self.versions.keys()) else self.versions['discover/view'] 
        query_parms = {'scope':scope.upper(), 'schema':schema.upper()}
        if version != 'v1' and search_pattern: query_parms['searchPattern'] = search_pattern
        success, rtn = self.call_api('discover/view',True,  SXTApiCallTypes.GET, query_parms=query_parms)
        return success, (rtn if success else [rtn]) 


    def discovery_get_columns(self, schema:str, table:str):
        """--------------------
        Connects to the Space and Time network and returns all available columns within a table.

        Calls and returns data from API: discover/table         
        
        Args:
            schema (str): Schema name for which to retrieve tables.
            table (str): Table name for which to retrieve columns.  This should be tablename only, NOT schema.tablename.

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list of dict. 
        """
        version = 'v2' if 'discover/table/column' not in list(self.versions.keys()) else self.versions['discover/table/column'] 
        schema_or_namespace = 'namespace' if version=='v1' else 'schema'
        query_parms = {schema_or_namespace:schema.upper(), 'table':table}
        success, rtn = self.call_api('discover/table/column',True,  SXTApiCallTypes.GET, query_parms=query_parms)
        return success, (rtn if success else [rtn]) 
    


    def subscription_set_name(self, name:str) -> tuple[bool, dict]:
        """--------------------
        Assigns a user-friendly name to an existing subscription.
        
        Args: 
            name (str): Subscription user-friendly name.  

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        if len(name)==0: return False, 'Name cannot be empty.'
        success, rtn = self.call_api('subscription/name', True, SXTApiCallTypes.PUT, query_parms={'subscriptionName':name})
        return success, {n:v for n,v in rtn.items() if v} if success else rtn


    def subscription_get_info(self):
        """--------------------
        Retrieves information on the authenticated user's subscription from the Space and Time network.

        Calls and returns data from API: subscription         
        
        Args: 
            None 

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        endpoint = 'subscription'
        version = 'v2' if endpoint not in list(self.versions.keys()) else self.versions[endpoint] 
        success, rtn = self.call_api(endpoint=endpoint, auth_header=True, request_type=SXTApiCallTypes.GET )
        return success, (rtn if success else [rtn]) 
     

    def subscription_get_users(self):
        """--------------------
        Retrieves information on all users of a subscription from the Space and Time network.  May be restricted to Admin or Owners.

        Calls and returns data from API: subscription/users         
        
        Args:
            None

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        endpoint = 'subscription/users'
        version = 'v2' if endpoint not in list(self.versions.keys()) else self.versions[endpoint] 
        success, rtn = self.call_api(endpoint=endpoint, auth_header=True, request_type=SXTApiCallTypes.GET )
        return success, (rtn if success else [rtn]) 
    

    def subscription_invite_user(self, role:str = 'member'):
        """--------------------
        Creates a subcription invite code (aka joincode).  Can join as member, admin, owner.

        Calls and returns data from API: subscription/invite.  
        Allows an Admin or Owner to generate a joincode for another user, who (after authenticating) 
        can consume the code and join the subcription at the specified level. 
        The code is only valid for 24 hours, and assigned role cannot be greater than the creator
        (i.e., an Admin cannot generate an Owner code).

        Args: 
            role (str): Role level to assign the new user. Can be member, admin, or owner.
        
        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        endpoint = 'subscription/invite'
        role = role.upper().strip()
        if role not in ['MEMBER','ADMIN','OWNER']:
            return False, {'error':'Invites must be either member, admin, or owner.  Permissions cannot exceed the invitor.'}
        version = 'v2' if endpoint not in list(self.versions.keys()) else self.versions[endpoint] 
        success, rtn = self.call_api(endpoint=endpoint, auth_header=True, request_type=SXTApiCallTypes.POST,
                                     query_parms={'role':role} )
        return success, (rtn if success else [rtn]) 


    def subscription_join(self, joincode:str):
        """--------------------
        Allows the authenticated user to join a subscription by using a valid joincode.

        Calls and returns data from API: subscription/invite/{joinCode}.  
        Note, joincodes are only valid for 24 hours.

        Args: 
            joincode (str): Code created by an admin to allow an authenticated user to join their subscription.
        
        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        endpoint = 'subscription/invite/{joinCode}'
        version = 'v2' if endpoint not in list(self.versions.keys()) else self.versions[endpoint] 
        success, rtn = self.call_api(endpoint=endpoint, auth_header=True, request_type=SXTApiCallTypes.POST,
                                     path_parms= {'joinCode': joincode} )
        return success, (rtn if success else [rtn]) 


    def subscription_leave(self):
        """--------------------
        Allows the authenticated user to leave their subscription.

        Calls and returns data from API: subscription/leave.  

        Args: 
            None
        
        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json). 
        """
        endpoint = 'subscription/leave'
        version = 'v2' if endpoint not in list(self.versions.keys()) else self.versions[endpoint] 
        success, rtn = self.call_api(endpoint=endpoint, auth_header=True, request_type=SXTApiCallTypes.POST )
        return success, (rtn if success else [rtn])


    def subscription_get_users(self) -> tuple[bool, dict]:
        """
        Returns a list of all users in the current subscription.

        Args:
            None

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Dictionary of User_IDs and User Permission level in the subscription, or error as json.
        """
        endpoint = 'subscription/users'
        version = 'v2' if endpoint not in list(self.versions.keys()) else self.versions[endpoint] 
        success, rtn = self.call_api(endpoint=endpoint, auth_header=True, request_type=SXTApiCallTypes.GET )
        if success: rtn = rtn['roleMap']
        return success, rtn
        

    def subscription_remove(self, User_ID_to_Remove:str) -> tuple[bool, dict]:
        """
        Removes another user from the current user's subscription.  Current user must have more authority than the targeted user to remove.

        Args: 
            User_ID_to_Remove (str): ID of the user to remove from the current user's subscription.

        Returns:
            bool: Success flag (True/False) indicating the api call worked as expected.
            object: Response information from the Space and Time network, as list or dict(json).
        """
        endpoint = 'subscription/remove/{userId}'
        version = 'v2' if endpoint not in list(self.versions.keys()) else self.versions[endpoint] 
        success, rtn = self.call_api(endpoint=endpoint, auth_header=True, request_type=SXTApiCallTypes.POST,
                                     path_parms= {'userId': User_ID_to_Remove} )
        return success, rtn
        


if __name__ == '__main__':

    token = 'eyJ0eXBlIjoiYWNjZXNzIiwia2lkIjoiZTUxNDVkYmQtZGNmYi00ZjI4LTg3NzItZjVmNjNlMzcwM2JlIiwiYWxnIjoiRVMyNTYifQ.eyJpYXQiOjE3Mzg3MTAxMDQsIm5iZiI6MTczODcxMDEwNCwiZXhwIjoxNzM4NzExNjA0LCJ0eXBlIjoiYWNjZXNzIiwidXNlciI6InN0ZXBoZW4iLCJzdWJzY3JpcHRpb24iOiIzMWNiMGI0Yi0xMjZlLTRlM2MtYTdhMS1lNWRmNDc4YTBjMDUiLCJzZXNzaW9uIjoiMjAwMTMzN2EyZDdmZmIxMGU1ZDJlM2Y1Iiwic3NuX2V4cCI6MTczODc5NjUwNDI3OSwiaXRlcmF0aW9uIjoiZWQ2MjhkYTU2NTVmZjQzZGU3OTMxODg1In0.NUkgbTcgUWLYJaJVoRrh_evNJlp0IiCYUGR3oV7_JILpg4g-UQO1ojsHutcGp5a72uBO5vHy75Nsq8s-VaEYgA'
    api = SXTBaseAPI(token)

    success, response = api.auth_code_register(user_id = 'poopsylicious', joincode='', email = 's@a.com')
    
    success, response = api.subscription_set_name('SXT Labs')

    print( api.subscription_get_info() )
    success, users = api.subscription_get_users() 

    success, response = api.subscription_invite_user(role='member')
    joincode = response['text']

    print( api.subscription_join(joincode=joincode) )

    pass 


import os, sys, pytest, pandas, random
from pathlib import Path

# load local copy of libraries
sys.path.append(str( Path(Path(__file__).parents[1] / 'src').resolve() ))
from spaceandtime.spaceandtime import SXTUser


def test_remove_all_users_from_test_subscription():
    # login with admin
    admin = SXTUser(dotenv_file='.env_loader_admin')
    admin.authenticate()
    assert admin.user_id == 'testuser_owner'

    # get list of all users in subscription
    success, users = admin.get_subscription_users()
    assert success
    for userid, role in users.items():
        if userid == admin.user_id: continue # skip self

        # remove all other users from subscription
        success, response = admin.remove_from_subscription(userid)
        assert success

    success, users = admin.get_subscription_users()
    assert len(users) == 1


def test_adding_users_to_subscription():
    # login with admin 
    steve = SXTUser(dotenv_file='.env_loader_admin')
    steve.authenticate()
    assert steve.user_id == 'testuser_owner'

    sxtloader_users = []

    # create N new load users
    for i in range(0,5):

        # load keys, then override name
        sxtloaderN = SXTUser(dotenv_file='.env_loader',
                            user_id=f'testuser_joincode{i}_{str(random.randint(0,999999)).zfill(6)}',)
        assert 'disconnected' in sxtloaderN.subscription_id
        assert sxtloaderN.exists == False
        
        # register new user and add to subscription
        joincode = steve.generate_joincode(role='member')
        assert len(joincode) > 0

        success, response = sxtloaderN.register_new_user() 
        assert sxtloaderN.exists == True
        assert sxtloaderN.subscription_id == ''
        first_access_token = sxtloaderN.access_token

        if success: success, response = sxtloaderN.join_subscription(joincode)
        assert sxtloaderN.subscription_id != ''
        assert 'disconnected' not in sxtloaderN.subscription_id
        assert first_access_token != sxtloaderN.access_token

        success, response = sxtloaderN.leave_subscription()
        

if __name__ == '__main__':
    test_remove_all_users_from_test_subscription()
    test_adding_users_to_subscription()
    pass 


import logging, base64
from pathlib import Path 
import nacl.signing
from biscuit_auth import KeyPair, PrivateKey 
from .sxtexceptions import SxTKeyEncodingError
from .sxtenums import SXTKeyEncodings



####
#### SXT KEY MANAGER
####
class SXTKeyManager():
    """Class to manage creation and maintenance of keys and biscuits."""

    biscuits:list = []
    logger:logging.Logger = None
    warning_for_biscuit_length = 1800
    keychange_callback_func_list = []
    __pv:bytes = bytes(''.encode())
    __pb:bytes = bytes(''.encode())
    __en:SXTKeyEncodings = SXTKeyEncodings.HEX
    ENCODINGS = SXTKeyEncodings


    def __init__(self, private_key:str = None, new_keypair: bool = False, encoding:SXTKeyEncodings = None, keychange_callback_func = None, logger:logging.Logger = None) -> None:
        """Class to manage creation and maintenance of keys and biscuits."""
        if logger: 
            self.logger = logger 
        else: 
            self.logger = logging.getLogger()
            self.logger.setLevel(logging.INFO)
            if len(self.logger.handlers) == 0: 
                self.logger.addHandler( logging.StreamHandler() )
        self.logger.info('new SXT KeyManager initiated')
        self.keychange_callback_func_list = []
        if keychange_callback_func: self.add_keychange_callback(keychange_callback_func)

        if encoding: self.encoding = encoding
        if new_keypair: 
            self.new_keypair()
            return None
        if private_key: self.private_key = private_key
        return None
    

    def __str__(self):
        flds = self.__keydict__()
        flds['private_key'] = flds['private_key'][:6]+'...'
        return '\n'.join( [ f'\t{n} = {v}' for n,v in flds.items() ] )
    
    def __repr__(self):
        return  '\n'.join( [ f'\t{n} = {v}' for n,v in self.__keydict__().items() ] )
    
    def __keydict__(self, keychanged:str = None) -> dict:
        rtn = {'private_key': self.private_key, 
                'public_key': self.public_key, 
                'encoding': self.encoding.name }
        if keychanged: rtn['key_changed'] = keychanged
        return rtn
    
    def __callback__(self, keychanged:str ) -> None:
        for func in self.keychange_callback_func_list:
            func( self.__keydict__(keychanged) )

    @property
    def private_key(self):
        return self.convert_key(self.__pv, SXTKeyEncodings.BYTES, self.encoding)
    @private_key.setter
    def private_key(self, value):
        self.__pv = self.convert_key(value, self.get_encoding_type(value), SXTKeyEncodings.BYTES) if value else ''
        self.__pb = ''
        self.__callback__('private_key')
        self.logger.debug(f'private key updated to { self.__pv[:6] }...')

    @property
    def public_key(self):
        if self.__pv and len(self.__pb)==0:
            kp = KeyPair.from_private_key(PrivateKey.from_hex( self.convert_key(self.__pv, encoding_out = SXTKeyEncodings.HEX)))
            self.__pb = bytes(kp.public_key.to_bytes())
            self.__callback__('public_key')
        return self.convert_key(self.__pb, encoding_out= self.encoding)
    @public_key.setter
    def public_key(self, value):
        self.__pb = self.convert_key(value, self.get_encoding_type(value), SXTKeyEncodings.BYTES) if value else ''
        self.__callback__('public_key')
        self.logger.debug(f'public key updated to { self.__pb }...')

    @property
    def encoding(self):
        return self.__en
    @encoding.setter
    def encoding(self, value) -> str:
        if not value in SXTKeyEncodings:
            raise SxTKeyEncodingError("Invalid encoding option, must be a member of SXTKeyEncodings", logger=self.logger)
        self.__en = value

    def private_key_to(self, encoding_out: SXTKeyEncodings = SXTKeyEncodings.HEX):
        return self.convert_key( key=self.__pv, encoding_out = encoding_out )
    
    def public_key_to(self, encoding_out: SXTKeyEncodings = SXTKeyEncodings.HEX):
        if len(self.__pb)==0: x = self.public_key # trigger property to refresh
        return self.convert_key( key=self.__pb, encoding_out = encoding_out )
    
    def get_encoding_type(self, key) -> str:
        """--------------------
        Accepts a key str or bytes, and returns the encoding type, [bytes, hex, base64].

        Args:
            key (any): Key to evaluate, as a string or bytes.

        Returns:
            str: Encoding type, [bytes, hex, base64]

        Examples:
            >>> SXTKeyManager().get_encoding_type("k6G2adpHxohA9sOBwHV8KRE5eDAJ/IEfocv5zkODgjA=")
            base64
            >>> SXTKeyManager().get_encoding_type("7063e65f0ba0e2aaaeb7d240248be19fea6f68dcccb50e0f2de3e22595f84751")
            hex
            >>> SXTKeyManager().get_encoding_type(b'\x93\xa1\xb6i\xdaG\xc6\x88@\xf6\xc3\x81\xc0u|)\x119x0\t\xfc\x81\x1f\xa1\xcb\xf9\xceC\x83\x820')
            bytes
        """
        if type(key) == bytes and len(key) == 32: return SXTKeyEncodings.BYTES
        try:
            bytes.fromhex(key)
            return SXTKeyEncodings.HEX
        except:
            if type(key) == str and len(key) == 44: return SXTKeyEncodings.BASE64
        raise SxTKeyEncodingError(f'Unknown Encoding: {key}', logger=self.logger)


    def new_keypair(self) -> dict:
        """--------------------
        Generate a new ED25519 keypair, set class variables and return dictionary of values.

        Returns: 
            dict: New keypair values

        Examples: 
            >>> km = SXTKeyManager(SXTKeyEncodings.BASE64)
            >>> km.new_keypair
            ['private_key', 'public_key']
            >>> len( km.private_key )
            64
            >>> km.encoding = SXTKeyEncodings.BASE64
            >>> len( km.private_key )
            44
         """
        keypair = KeyPair()
        self.private_key = bytes(keypair.private_key.to_bytes())
        return { 'private_key': self.private_key
                ,'public_key':  self.public_key }


    def convert_key(self, key, encoding_in:SXTKeyEncodings = SXTKeyEncodings.BYTES
                             , encoding_out:SXTKeyEncodings = SXTKeyEncodings.HEX):
        """--------------------
        Converts a key value from one stated format into requested encoding format.

        Args: 
            key (any): Key value, typically either str [base64, hex] or bytes.
            encoding_in (str):  Encoding of supplied key, as SXTKeyEncodings
            encoding_out (str): Encoding of returned key, as SXTKeyEncodings
        
        Return: 
            dict: Converted key the encoding_out encoding.

        Examples:
            >>> SXTKeyManager().convert_key('0123456789abcdef', SXTKeyEncodings.HEX, SXTKeyEncodings.BASE64)
            ASNFZ4mrze8=
            >>> SXTKeyManager().convert_key('ASNFZ4mrze8=', SXTKeyEncodings.BASE64, SXTKeyEncodings.HEX)
            0123456789abcdef
        """
        try:
            # always take to bytes first
            if not key:
                key_bytes = bytes(b'')
            elif encoding_in == SXTKeyEncodings.BYTES:
                key_bytes = bytes(key)
            elif encoding_in == SXTKeyEncodings.BASE64:
                key_bytes = base64.b64decode(key)
            elif encoding_in == SXTKeyEncodings.HEX:
                key_bytes = bytes.fromhex(key)

            # format as requested encoding
            if encoding_out == SXTKeyEncodings.BYTES:
                key_out = key_bytes
            elif encoding_out == SXTKeyEncodings.BASE64:
                key_out = base64.b64encode(key_bytes).decode('utf-8')
            elif encoding_out == SXTKeyEncodings.HEX:
                key_out = key_bytes.hex()

            # self.logger.debug(f'Key verified and converted from {encoding_in.name} to {encoding_out.name}.')
            return key_out
        except Exception as ex:
            error = ex
        raise SxTKeyEncodingError(f'Error: {error}, going from {encoding_in.name} to {encoding_out.name}', logger=self.logger)


    def get_KeyPair(self) ->KeyPair:
        """Builds and returns a KeyPair object from current private / public key."""
        if not self.__pv: 
            raise ValueError('Requires valid private_key to be set')
        kp = KeyPair.from_private_key( PrivateKey.from_bytes(self.__pv) )
        return kp


    def sign_message(self, message:str, encoding_out:SXTKeyEncodings = SXTKeyEncodings.HEX):
        """--------------------
        Use private key to cryptographically sign and return message.

        Args: 
            message (str): String message to sign with the class private key and return.
            encoding_out (SXTKeyEncodings): Encoding of returned signed message, as SXTKeyEncodings
            
        Returns:
            str | bytes: Signed message, encoded per encoded_out (or class.encoding as default)

        """
        if type(message)!=str: 
            raise ValueError(f'paramter: "message" must be a string type, not {str(type(message))}')
        try:
            if not encoding_out: encoding_out = self.encoding
            signing_object = nacl.signing.SigningKey(bytes(self.__pv))
            signed_message = signing_object.sign(message.encode('utf-8'))
            return self.convert_key(signed_message.signature, SXTKeyEncodings.BYTES, encoding_out)
        except Exception as ex:
            error = ex
        raise SxTKeyEncodingError(error, logger=self.logger)
        

    def add_keychange_callback(self, func) -> None:
        """Adds a function to a list of functions to call whenever a key (public or private) changes."""
        if type(self.keychange_callback_func_list) != list: self.keychange_callback_func_list = []
        self.keychange_callback_func_list.append(func)


    def clear_keychange_callback(self) -> None:
        """Clears all functions from the keychange callback list."""
        self.keychange_callback_func_list = []



if __name__ == '__main__':

    print('\n', '-=-='*10, '\n' )

    # BASIC USAGE
    key1 = SXTKeyManager(new_keypair=True)
    key1.new_keypair()  # or, explicitly

    # objects come with default root logger with debug msgs to stdout
    key1.logger.debug( key1 )
    print( 'signed message:', key1.sign_message('my signed message') )



    # Supply your own setup to the constructor
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    if len(logger.handlers) == 0: logger.addHandler( logging.StreamHandler() )
    
    key2 = SXTKeyManager(private_key = '4G7l7Zu4zTTMsV/p3i7qjNEqf2LV92LSXfntQXWkH+c=', 
                         encoding = SXTKeyEncodings.HEX, 
                         keychange_callback_func = print, # function called on key-change
                         logger=logger
                         )
    # during instantiation, private key is set, and that change triggers the callback 'print'
    # then the public key is regenerated, again triggering the callback 'print'.
    # The two outputs have different 'key_changed' values: one private_key, one public_key.

    # casting or printing as a string (__str__) provides a human-readable string output:
    print( key2 , '\n' + '-'*10 )

    # to print the object without private key truncation, try the repr() function:
    print( repr(key2) )

     

    # you can have many callback functions, which fire on any key value change
    import pprint
    key2.clear_keychange_callback()
    key2.encoding = SXTKeyEncodings.BASE64

    print( '-'*10, 'adding callbacks...')
    key2.add_keychange_callback( print ) 
    key2.add_keychange_callback( pprint.pprint )
    key2.add_keychange_callback( lambda x: print(f"\tKey that Changed: {x['key_changed']}") )

    # triggers 6 callbacks: 3 for the private key change, which triggers 3 more with a public key rebuild
    key2.private_key = 'mOXd6UzMNiR1qfQYW/MRcVANdxP+k62JGYXo2t5Ukv0='
    
    # this only happens on change, not every time the item is inspected.
    print( key2.public_key )
    print( key2.private_key )
    

    # the encoding property only controls the encoding OUTPUT.  
    # All encoding INPUT is automatically detected and stored as bytes (note debug message)
    key2.clear_keychange_callback()
    key2.add_keychange_callback( lambda x: print('\t', x['private_key'] ) )
    key2.logger.setLevel( logging.INFO )

    for encoding in [ SXTKeyEncodings.HEX, SXTKeyEncodings.BASE64, SXTKeyEncodings.BYTES]:
        key2.encoding = encoding
        print( encoding.name )
        key2.private_key = '4G7l7Zu4zTTMsV/p3i7qjNEqf2LV92LSXfntQXWkH+c='
        key2.private_key = 'e06ee5ed9bb8cd34ccb15fe9de2eea8cd12a7f62d5f762d25df9ed4175a41fe7'
        key2.private_key = b'\xe0n\xe5\xed\x9b\xb8\xcd4\xcc\xb1_\xe9\xde.\xea\x8c\xd1*\x7fb\xd5\xf7b\xd2]\xf9\xedAu\xa4\x1f\xe7'


    # Access the native KeyPair object, if needed:
    key2.encoding = SXTKeyEncodings.BYTES
    privatekey_obj = key2.get_KeyPair().private_key.from_bytes( key2.private_key )
    print( privatekey_obj.to_hex() )
    

    # note: key properties will convert keys from BYTES to requested encoding 
    # upon inspection - even by the IDE.   Try hover-over inspecting in an IDE
    # and notice the debug log entries.
    key2.logger.setLevel( logging.DEBUG )
    key2.encoding = SXTKeyEncodings.BASE64
    key2.private_key
    key2.public_key

    pass


import os, sys, pytest, pandas, random
from pathlib import Path

# load local copy of libraries
sys.path.append(str( Path(Path(__file__).parents[1] / 'src').resolve() ))
from spaceandtime.spaceandtime import SpaceAndTime
from spaceandtime.spaceandtime import SXTUser
API_URL = 'https://api.spaceandtime.dev'




# create test file:
def create_random_test_file(skip_pos: int = None):
    rndnum = random.randint(1,1e9)
    filepaths = []
    for iloop, testsetup in enumerate( ['USER', 'SXT_USER', 'USER_', 'SXT_USER_'] ):
        filepath = Path(f'./tests/user_saves/{testsetup}--test.{rndnum}.env')
        filecontent = [f'API_URL=https://api.spaceandtime.dev\n']
        if iloop != skip_pos: 
            filecontent.append(f'{testsetup}ID=testuser_{rndnum}\n')
            filecontent.append(f'{testsetup}_PRIVATE_KEY=Z833BwZcwotJf4zVA89HlyvxH8xqAUOXzTcR1dWhsrk=\n')
            filecontent.append(f'{testsetup}_PUBLIC_KEY=Z833BwZcwotJf4zVA89HlyvxH8xqAUOXzTcR1dWhsrk=\n')
        filecontent = str(''.join(filecontent)).replace('__','_')
        with open(filepath, 'w') as f:
            f.write(filecontent)
        filepaths.append(filepath)
    return rndnum, filepaths


    
def test_random_user_envars():

    for skiptest in [None, 0, 1, 2, 3]:

        # create new login .env file
        rndnum, envfilepaths = create_random_test_file(skiptest)

        # load new .env file 
        for iloop, envfilepath in enumerate(envfilepaths):
            if iloop == skiptest: continue
            sxt = None 
            os.environ.clear()
            with open(envfilepath, 'r') as f:
                print(f.read())
            sxt = SpaceAndTime(envfile_filepath = envfilepath)
            assert sxt.user.user_id == f'testuser_{rndnum}'
            assert sxt.user.private_key == 'Z833BwZcwotJf4zVA89HlyvxH8xqAUOXzTcR1dWhsrk='
            

 
test_random_user_envars()



import sys,random, time
from pathlib import Path
from datetime import datetime

# load local copy of libraries
sys.path.append(str( Path(Path(__file__).parents[1] / 'src').resolve() ))
from spaceandtime.spaceandtime import SpaceAndTime
from spaceandtime.sxtkeymanager import SXTKeyManager
from spaceandtime.sxtbiscuits import SXTBiscuit
from spaceandtime.sxtresource import SXTTable, SXTView, SXTMaterializedView


# authenticate once for all subsequent tests
envfile = Path(Path(__file__).parent / '.env').resolve() # ./tests/.env
sxt = SpaceAndTime(envfile_filepath=envfile, application_name='pytest_SDKTesting')
sxt.logger_addFileHandler( Path(envfile.parent / 'latest_test_log.txt') )
sxt.authenticate()


def test_all():
    num = '{num:06d}'.format(num=random.randint(0,999999))
    biscuit_file = Path(Path(__file__).parent / 'backup_biscuits' / f"{datetime.now().strftime('%Y%m%d.%h%m%s')}_biscuits_from_tests.json")

    try:
        # create one keypair / biscuit for all use-cases and objects 
        biscuit = SXTBiscuit('all_uses', new_keypair=True)
        biscuit.add_capability('*', sxt.GRANT.ALL)
        biscuit.save(biscuit_file) # in case something goes wrong... we can rm at the end

        # Create table
        table = SXTTable(f'SXTTest.MyTestTable_{num}', key_manager=biscuit.key_manager, access_type=sxt.TABLE_ACCESS.PUBLIC_READ, SpaceAndTime_parent=sxt)
        table.add_biscuit_object(biscuit)
        table.create_ddl = table.create_ddl_sample
        table_success, resopnse = table.create()
        assert table_success
        assert table.exists

        # load some made-up data
        data = [{'MyID':n, 'MyName':chr(n), 'MyDate':'2024-01-01'} for n in range(65,65+26)]
        insert_success, response = table.insert.with_list_of_dicts(data)
        assert insert_success

        # Select, using built-in function
        select_success, data = table.select() 
        assert select_success
        assert len(data) == 26
        assert data[0]['MYDATE'][:10] == '2024-01-01'
        assert 64 < data[0]['MYID'] < 90
        assert sorted([d['MYID'] for d in data])[0] == 65 # sorted here, in python

        # Select, using custom SQL
        select_success, data = table.select(f'Select * from {table.table_name} order by MyID')
        assert select_success
        assert len(data) == 26
        assert data[0]['MYID'] == 65 # sorted in the SQL

        # create view on table
        view = SXTView(f'SXTTest.MyTestView_{num}_Evens', key_manager=biscuit.key_manager, SpaceAndTime_parent=sxt)
        view.add_biscuit_object(biscuit)
        view.create_ddl = f"""
            CREATE VIEW {view.view_name} 
            {view.with_statement} AS
            SELECT * FROM {table.table_name} WHERE MyID % 2 = 0 """
        view_success, response = view.create()
        assert view_success
        assert view.exists
        
        # views can take a moment for network to sync sometimes, let's check a few times
        for i in range(1,30):
            select_success, data = view.select()
            if select_success and len(data) == 13: break
            time.sleep(10) 

        # select from the EVENS view (only even numbers)
        assert select_success
        assert len(data) == 13
        assert sorted([d['MYID'] for d in data])[0] == 66

        # create materialized view
        matview = SXTMaterializedView(f'SXTTest.MyTestMatView_{num}_Odds', key_manager=biscuit.key_manager, SpaceAndTime_parent=sxt)
        matview.add_biscuit_object(biscuit)
        matview.create_ddl_template = f"""
            CREATE VIEW {matview.matview_name} 
            {matview.with_statement} AS
            SELECT * FROM {table.table_name} 
            WHERE MyID not in ( Select MyID from {view.view_name} ) """
        view_success, response = matview.create()
        assert view_success
        assert matview.exists

        for i in range(1,30):
            select_success, data = matview.select()
            if select_success and len(data) == 13: break
            time.sleep(10) 

        # select from the ODDS materialized view (only odd numbers)
        assert select_success
        assert len(data) == 13
        assert sorted([d['MYID'] for d in data])[0] == 65

    
    except Exception as ex:
        # for any error, just try to drop everything
        pass 

    # make sure to drop in reverse order of dependencies
    dropmatview_success = dropview_success = droptable_success = True
    if 'matview' in locals(): dropmatview_success, response = matview.drop()
    if 'view' in locals(): dropview_success, response = view.drop()
    if 'table' in locals(): droptable_success, response = table.drop()
    
    # if anything created fails to drop, hang onto the biscuit (or if )
    if (dropmatview_success and dropview_success and droptable_success):
        Path(biscuit_file).unlink(True)



if __name__ == '__main__':
    test_all()
    pass


from .spaceandtime import SpaceAndTime
from .sxtbaseapi import SXTBaseAPI
from .sxtbiscuits import SXTBiscuit
from .sxtkeymanager import SXTKeyManager
from .sxtresource import SXTResource, SXTTable, SXTView, SXTMaterializedView
from .sxtuser import SXTUser
from .sxtenums import *
from .sxtexceptions import *

 